/********************************************************************/
/*                                                                  */
/*   Program Name :  pulsar                                         */
/*                                                                  */
/*      Version:  0.00  2020-05-26                                  */
/*                                                                  */
/*      Copyright (c) 2020-     T.Kondo      All Right Reserved     */
/*                                                                  */
/********************************************************************/
/*
 
 display pulsar with ensemble average of power with period given from an operator
 for VSSP data format

        Ver. 2020-05-26 start development 

    How to run:
       pulsar vssp_file pulse_period [options]

        where 
               vssp_file -- VSSP32 format data file name
               pulse_period  ---  pulsar period (sec)

         [options (any order)]

              -t[length]  tlength -- processing data period (sec)
                                  default all
              -ref[time]  hh:mm:ss  -- set reference time
              -prt        hh:mm:ss  -- set reference time

              -p[mode] pmode
                    plot device selection
                        pmode=0 :  display and PostScript file out
                             =1 :  PostScript file out only
                             =2 :  display only
                             default is 2
              -o[ffset] soffset -- start time offset
              -ps   --  set postscript out
              -h[alt] --  halt after each plot (default continuous plot) 
              -s[msec] smsec
                         set suspend time (msec) after each plot
                         default is 0 (as soon as posiible base)
              -power --  set plot power not amp   (added on 2020-05-20)
              -min[pow] minpow  ---  set minimum power level in a graph (valid for power mode)
              -max[pow] maxpow  ---  set maximum power level in a graph (valid for power mode)
              -tave tave  --- averaging period (time constant) for power mode 2020-05-20
*/
#define PROGRAM_NAME "pulsar"
#define PROGRAM_VERSION "Ver. 2020-05-26"

#include "pulsar.h"    // header for graphics

#include "g_head.h"    // header for graphics

#ifdef WIN32
  #define USE_FFTW    // use FFTW package
#endif

#ifdef USE_FFTW
 #ifdef WIN32
  #include "fftw3.h"     // use of FFTW package
 #else
  #include <fftw3.h>     // use of FFTW package
 #endif
#endif

#include "libk5.h"      // common header for IP-VLBI utilities

#include "libfft.h"

#ifdef WIN32
#define USE_ASM    // use inline assemble function
#endif

#define DEFLT_DATA_FILE "tds.data"   // default data file name

//#define DEFLT_SLEEP_MSEC 100        // default sleep suspend time in msec
#define DEFLT_D_POINTS 100     // default display span in # of data

#define MOVING_AVE        // take moving average in case of power mode

void print_usage(speana_param_t *st)
{
    char pgdevice_dflt[20];
    printf("%s %s",PROGRAM_NAME, PROGRAM_VERSION);
#ifdef USE_GNUPLOT   /* using gnuplot */
    printf(" compliled for GNUPLOT \n");
    strcpy(pgdevice_dflt,G_DISPLAY);
#else
    printf(" compliled for PGPLOT \n");
    strcpy(pgdevice_dflt,st->g_display);
#endif
    printf("\n");
    printf("Usage: %s vsspfile pulse_period [options]\n", PROGRAM_NAME);
    printf("\n");
    printf("         where  vsspfile ---- VSSP Data file name\n");
    printf("                pulse_period ---- pulsar period (sec)\n");
    //printf("                              if 0 is given, tds.data will be used\n");
    printf("\n");
    printf("            [options]\n");
//    printf("                -p[mode] pmode\n");
//    printf("                     plot device selection\n");
//    printf("                        pmode=0 :  display and PostScript file (%s) out\n",st->ps_fname);
//    printf("                             =1 :  PostScript file (%s) out only\n",st->ps_fname);
//    printf("                             =2 :  display only\n");
//    printf("                            default is 2\n");
    printf("                -t[length] tlength --- processing data period (sec)\n");
    printf("                -ref[time] hh:mm:ss  -- set reference time\n");
        //printf("                -c comment --- Comment appeared on a graph\n");
    //printf("                -ps  --- post script out too\n");
/*
    printf("                -h[alt]  --- halt after each plot (default continuous plot)\n");
    printf("                -s[msec] smsec\n");
    printf("                     set suspend time (msec) after each plot\n");
    printf("                     default is 0 (as soon as possible base)\n");
    printf("                -power   --- set power plot mode\n");
    printf("                -min[pow] minpow -- set minimum power level in graph (valid for power mode)\n");
    printf("                -max[pow] maxpow -- set maximum power level in graph (valid for power mode)\n");
    printf("                -tave tave --- averaging period (time constant) for power mode\n");

#ifndef USE_GNUPLOT   // using gnuplot
    printf("\n");
    printf("  Note: %s ---  environment variable for PGPLOT display device\n",PGDISP);
    printf("                   = %s",getenv(PGDISP));
    if(getenv(PGDISP)==NULL) {
        printf(",  program default is ( %s )\n",PGDISP_DFLT);
    } else {
        printf("\n");
    }
#endif
*/
    printf("\n");
    exit(0);
}

void runparam_get(int argc, char *argv[],runp_t *runp)                       
{
    /* get run parameters type 1 (conventional run tyep) */

    int n;
    char cbuf[256];
    char sbuf[10];

	// initialize parameters

    runp->fname[0]='\0';
    runp->comment[0]='\0';
    runp->soffset=0.0;
    runp->tspan=0.0;
    runp->pmode=2;
    runp->flag_halt=0;
    //runp->sleep_msec=DEFLT_SLEEP_MSEC;
    runp->sleep_msec=0;
    runp->flag_power = 1;  // 2020-05-26
    runp->minpow = 0.0;  // 2020-05-20
    runp->maxpow = 0.0;  // 2020-05-20
    runp->tave = 0.0;    // 2020-05-20
    runp->period = 0.0;
    runp->phase = 0.0;
    for (n = 0; n < 5; n++) runp->iprt[n] = -1;
    runp->tlength = 0.0;


    if (argc>=2){
        strcpy(runp->fname,*(argv+1));
    }
    if (strlen(*(argv+1))==1 && runp->fname[0]=='0'){
          strcpy(runp->fname, DEFLT_DATA_FILE);
    }

    if (argc>=3){
        runp->period=atof(*(argv+2));
    }
    
    //if(cmnd_get(argc,argv,"-p",cbuf)==0) runp->pmode = atoi(cbuf);     // -pmode 
    if(cmnd_get_np(argc,argv,"-p",cbuf)==0) runp->pmode = atoi(cbuf);     // -pmode 
    if(cmnd_get_np(argc,argv,"-pmode",cbuf)==0) runp->pmode = atoi(cbuf);     // -pmode   2020-05-20
    if(cmnd_get_np(argc,argv,"-t",cbuf)==0) runp->tlength = atof(cbuf);     //  -tlength 
    if(cmnd_get(argc,argv,"-tspan",cbuf)==0) runp->tspan = atof(cbuf);     //  -tspan 
    if(cmnd_get(argc,argv,"-a",cbuf)==0) runp->tspan = atof(cbuf);     //  -tspan   to keep compatibility with speana_n
    if(cmnd_get(argc,argv,"-i",cbuf)==0) runp->tspan = atof(cbuf);     //  -tspan   to keep compatibility with speana_n
    if(cmnd_get(argc,argv,"-c",cbuf)==0) strcpy(runp->comment,cbuf);   //  -comment 
    if(cmnd_get(argc,argv,"-o",cbuf)==0) runp->soffset = atof(cbuf);   //  -offset 
    //if(cmnd_get(argc,argv,"-ps",cbuf)==0) runp->pmode=0;               // -ps
    if(cmnd_get(argc,argv,"-h",cbuf)==0) runp->flag_halt=1;           // -halt
    if(cmnd_get(argc,argv,"-s",cbuf)==0) runp->sleep_msec = atoi(cbuf);   //  -sleep_msec 
    if(cmnd_get_np(argc,argv,"-power",cbuf)==0) runp->flag_power=1;  // -power   2020-05-20
    if(cmnd_get(argc,argv,"-min",cbuf)==0) runp->minpow=atof(cbuf);  // -minpower   2020-05-20
    if(cmnd_get(argc,argv,"-max",cbuf)==0) runp->maxpow=atof(cbuf);  // -minpower   2020-05-20
    if(cmnd_get(argc,argv,"-tave",cbuf)==0) runp->tave=atof(cbuf);  // -tave   2020-05-20
    if (cmnd_get(argc, argv, "-ref", cbuf) == 0) {  // set reference time
        strncpy(sbuf, cbuf, 2); sbuf[2] = '\0';
        runp->iprt[2] = atoi(sbuf);
        strncpy(sbuf, cbuf+3, 2); sbuf[2] = '\0';
        runp->iprt[3] = atoi(sbuf);
        strncpy(sbuf, cbuf+6, 2); sbuf[2] = '\0';
        runp->iprt[4] = atoi(sbuf);
    }
    if (cmnd_get(argc, argv, "-prt", cbuf) == 0) {  // set reference time
        strncpy(sbuf, cbuf, 2); sbuf[2] = '\0';
        runp->iprt[2] = atoi(sbuf);
        strncpy(sbuf, cbuf+3, 2); sbuf[2] = '\0';
        runp->iprt[3] = atoi(sbuf);
        strncpy(sbuf, cbuf+6, 2); sbuf[2] = '\0';
        runp->iprt[4] = atoi(sbuf);
    }

}


void calc_power(float *adata, int ch, int nadata, int nave, int mode)
{
    //  mode   1: running average   0: simple average

    int n,k;
    int nsum;
    int npos, nkai;
    int mave;
    double sum, ave;
    double p;
    float *wdata;

    // average calc
    sum = 0.0;
    nsum = 0;
    for (n = 0; n < nadata; n++) {
        sum += adata[n * 4 + ch];
        nsum++;
    }
    ave = sum / (double)nsum;
    // power calc
    for (n = 0; n < nadata; n++) {
        p = (adata[n * 4 + ch] - ave)*(adata[n * 4 + ch] - ave);
        adata[n * 4 + ch] = (float)p;
    }

    if (mode == 1) goto moving_ave;


    // power average calc
    nkai = nadata / nave + 1;

    for (n = 0; n < nkai; n++) {
        sum = 0.0;
        ave = 0.0;
        nsum = 0;
        for (k = 0; k < nave; k++) {
            npos = nave * n + k;
            if (npos >= nadata) break;
            sum += adata[npos * 4 + ch];
            nsum++;
        }
        if (nsum > 0) {
            ave = sum / (double)nsum;
        }
        else {
            ave = 0.0;
        }
        // set average data
        for (k = 0; k < nave; k++) {
            npos = nave * n + k;
            if (npos >= nadata) break;
            adata[npos * 4 + ch] = ave;
        }
    }
    goto exit2return;
moving_ave:;

    wdata = (float *)malloc(nadata * sizeof(float));
    for (n = 0; n < nadata; n++) wdata[n] = adata[n * 4 + ch];  // copy to wdata
    mave = nave;
    if ((nave % 2) == 0) mave++;    // set mave to odd number
    for (n = 0; n <= nadata - mave; n++) {
        npos = n + mave / 2;
        if (n == 0) {
            sum = 0.0;
            for (k = n; k < n + mave; k++) {
                sum += wdata[k];
            }
        }
        else {
            sum = sum - wdata[n - 1] + wdata[n + mave - 1];
        }
        ave = sum / (double)mave;
        adata[npos * 4 + ch] = ave;
        // edge process
        if (n == 0) {
            for(k=0; k< npos; k++) adata[k * 4 + ch] = ave;
        }
        if (n == nadata-mave) {
            for(k=npos+1; k< nadata; k++) adata[k * 4 + ch] = ave;
        }
    }
    free(wdata);
exit2return:;

}



int main(int argc, char *argv[])
{
    //int nloop,sekibun;
    char cw;
    int nloop;
    int flag_last;   // 2015-10-27
    int i,imax, m, n, k, kaisu;
    int npos;
    int npulse;
    int smode,usampl,nfft;
	int kbytes;
    int64_t bits, bytes;  // 2105-10-23
    int64_t ns_bopp;      // sample# from the 1st point of data file at begining of tsapn period
    int64_t ns_all;       // sample# from the 1st point of data file
    int numb;
    int p_usampl;     // current sample # in USAMPL
    int p_adata;      // current sample # in ADATA
    int flag_plot_done; 
    int nave;  // 2020-05-20
    //int scount[4];
    float fsampl,tsampl,fact;
    int isoffset;     // start offset in integer second      2014-08-12
    double fsoffset;  // fractional part of start offset     2014-08-12
    double ftime;     // time in second to judge start time  2014-08-12
    int start_flag;   // flag to judge start time            2014-08-12
    double nfact;
    double tinteg,tusampl;
    double tintegold;
    double telapsed;           // to measure processing time
    double dw;
    double tsec1st, tseclast;
    char lt1st[20], ltlast[20];
    clock_t cstart1sec, cfinish1sec;  // to calculate processing time
    char lprt[60];
    char cbuf[20];
    K5VSSP_file_t K5;
    speana_param_t st;
    runp_t runp;

    /* set graphic device */
    strcpy(st.ps_fname,PS_OUT_FILE);
#ifdef USE_GNUPLOT   // using gnuplot
    strcpy(st.g_display,G_DISPLAY);
    strcpy(st.g_ps_device,G_PS_DEVICE);
#else
    strcpy(st.g_display,select_env_dflt(PGDISP, PGDISP_DFLT));
    sprintf(st.g_ps_device,"%s%s",st.ps_fname,G_PS_DEVICE);
#endif

    if(argc == 1 || (argc == 2 && strncmp(*(argv+1),"--help",6)==0 )){
         print_usage(&st);
         exit(0);
    }
    if(argc == 2 && strncmp(*(argv+1),"--ver",5)==0 ){ 
          printf("pulsar %s",PROGRAM_VERSION);
#ifdef USE_GNUPLOT   /* using gnuplot */
          printf(" compliled for GNUPLOT \n");
#else
          printf(" compliled for PGPLOT \n");
#endif
          printf("\n");
          exit(0);
    }
    
    /* get run parameters */
    runparam_get(argc, argv, &runp);

    /* copy runparameters to other variables */
    strcpy(st.fname,runp.fname);
    strcpy(K5.fname,st.fname);

    strcpy(st.comment,runp.comment);
    st.soffset=runp.soffset;
    st.tspan=runp.tspan;
    st.toffset=runp.soffset;

    isoffset=(int)st.soffset;
    fsoffset=fmod(st.soffset,1.0);
    if(fsoffset > 0.0){
        start_flag=0;
    } else {
        start_flag=1;
    }
    ftime=0.0;
    st.flag_power = runp.flag_power;
    st.tave = runp.tave;


    printf("pulsar %s",PROGRAM_VERSION);
    printf("\n");

    cstart1sec = clock();

    printf("Data File is %s\n",K5.fname);
    K5.size=get_file_size(K5.fname);
    if( (K5.stream  = fopen( K5.fname, "rb" )) == NULL ){
         printf("File (%s) Open Error! \n",K5.fname);
        goto exitend2;
    }
    k5vssp_t_ini(&K5);


    /* Data header check to get sampling parameters */

    if((kbytes=CheckK5header_with_Read(&K5)) < 0){
        printf("File read error! \n");
        i=fclose( K5.stream ) ;
        goto exitend;
    }

    /* parameter set */

    st.pmode=runp.pmode;
    st.flag_halt=runp.flag_halt;
    st.sleep_msec=runp.sleep_msec;

    flag_last=0;
    st.sfreq=K5.sfreq;
    st.adbit=K5.adbit;
    st.numch=K5.numch;
    fsampl=(float)(1000.0*(float)st.sfreq);   // sampling frequency (Hz)
    tsampl=(float)(1.0/fsampl);            // sampling period (sec)
    bits=((int64_t)fsampl)*st.adbit*st.numch;  // # of bits in 1 sec data except header  2015-10-23
    bytes=bits/8;                     // # of bytes in 1 sec data except header
    K5.fsample=fsampl;
    //sekibun=1;                       // Averaging period for Spectrum (sec)
    K5.bytesec=(int)(bits/8);                  //# of bytes in 1 sec data

    //runp.dtime1st = K5.hh*3600.0 + K5.mm*60.0 + K5.ss*1.0;
    runp.dtime1st = (double)K5.seconds;
    runp.prtsec = runp.dtime1st;
    //runp.npdat = (int)(runp.period*fsampl + 0.5);
    runp.npdat = 1000;
    runp.pdat = (double *)malloc(runp.npdat * sizeof(double));
    runp.cdat = (int *)malloc(runp.npdat * sizeof(int));
    for (n = 0; n < runp.npdat; n++) {
        runp.pdat[n] = 0.0;
        runp.cdat[n] = 0;
    }

    smode=1;                         // low resolution mode

    for(m=0; m<st.numch; m++) st.chtable[m]=m+1;  // 2017-12-13

    st.minpow = runp.minpow;
    st.maxpow = runp.maxpow;
    //if (st.flag_power == 1) {
    //    st.minpow = 0.85;
    //    st.maxpow = 1.0;
    //}

    /* unit sample # set to read */

    if(smode ==1){ 
        usampl=10000; 
    } else {
        usampl=50000;
        if(st.sfreq == 40)  usampl=40000; 
        if(st.sfreq == 100) usampl=100000; 
        if(st.sfreq == 200) usampl=200000; 
    }
    imax=(st.sfreq*1000)/usampl;
    numb=usampl*st.adbit*st.numch/8;   // # of bytes in unit # of samples
    K5.numb=usampl*K5.adbit*K5.numch/8;       // # of bytes in a usampl
    K5.imax=(K5.sfreq*1000)/usampl;        // # of usampls in a second


    tusampl=(double)usampl/K5.fsample;   // period of usampl

    printf("imax numb  %d %d\n",K5.imax,K5.numb);    //debug


    /* set display time span default */
    if(st.tspan==0.0){
        //st.tspan=tusampl;
        st.tspan=DEFLT_D_POINTS*tsampl;
//        printf("display span set to %g(sec)\n",st.tspan);
    }

    /* set # of samples in tspan */
    dw=st.tspan*K5.fsample;
    if(dw > pow(2.0,31.0)/4.0){  //  bigger than maximum of 32-bit integer
        printf("given display span (%e sec) exceeds program limit %f\n",st.tspan,(pow(2.0,31.0)/4.0)/K5.fsample);
        printf("change parameter then run again!\n");
        i=fclose( K5.stream ) ;
        goto exitend;
    }
    st.nadata=(int)(dw+0.5);


    /* set FFT size */
    // not used in this program but to keep consistency with other programs

    nfft=1024;
    if(usampl < 524288) nfft= 524288;
    if(usampl < 262144) nfft= 262144;
    if(usampl < 131072) nfft= 131072;
    if(usampl < 65536)  nfft= 65536;
    if(usampl < 32768)  nfft= 32768;
    if(usampl < 16384)  nfft= 16384;

    st.nfft=nfft;
    st.usampl=usampl;

    nfact=2.0/(double)nfft;

    K5.nword=K5.numb/4;                 // # of 32bit-word in a usampl
    K5.tsample=1.0/K5.fsample;    // 2015-10-23
    K5.unitp=(double)usampl*K5.tsample;   // period in a usampl (sec)

    K5.ubuf  = (u_int32_t *)malloc(K5.nword*sizeof(u_int32_t)); // K5 data read buffer
    K5.idata  = (int *)malloc(32*K5.nword*sizeof(int)); // K5 data read buffer

    //st.xdata = (float *)malloc(nfft*sizeof(float));
    //st.adata = (float *)malloc(4*nfft*sizeof(float));
    //st.adata = (float *)malloc(4*usampl*sizeof(float));
    st.adata = (float *)malloc(4*st.nadata*sizeof(float));

    // set nave  2020-05-20
    if (runp.tave == 0.0) {
        nave = st.nfft;
        st.tave = (double)nave / fsampl;
    }
    else {
        nave = st.tave*fsampl;
    }


    /* foward startffset seconds */
    //if(fwdNsecK5(st.soffset, &K5) <0){
    if(fwdNsecK5(isoffset, &K5) <0){ 
        //printf("Foward error %d\n",st.soffset);
        printf("Foward error %d\n",isoffset);
        i=fclose( K5.stream ) ;
        goto exitend;
    }

//    printf("  Time %02d:%02d:%02d   Sec in Day = %d\n",
//        K5.hh,K5.mm,K5.ss,K5.seconds);
    runp.dtime = (double)K5.seconds;
    tsec1st = runp.dtime;
    tseclast = tsec1st;
    strcpy(lt1st, sectohhmmss(tsec1st));

    // set iprt if not set

    if (runp.iprt[2] < 0) {
        runp.iprt[2] = K5.hh;
        runp.iprt[3] = K5.mm;
        runp.iprt[4] = K5.ss;
    }
    runp.prtsec = runp.iprt[2] * 3600.0 + runp.iprt[3] * 60.0 + runp.iprt[4] * 1.0;

    if(K5.format==2){  // VSSP32          2014-06/13
	   sprintf(lprt,"PRT(UT) %04d/%03d %02d:%02d:%02d",
           K5.year,K5.totalday,runp.iprt[2],runp.iprt[3],runp.iprt[4]);
       runp.iprt[0] = K5.year;
       runp.iprt[1] = K5.totalday;
    } else {
       sprintf(lprt,"PRT(UT) %02d:%02d:%02d", runp.iprt[2], runp.iprt[3], runp.iprt[4]);
    }

   // if(fsoffset > 0.0){   // 2014-08-12
   //     sprintf(cbuf,".%03d",(int)(1000.0*fsoffset));
   //     strcat(lprt,cbuf);
   // }
    strcpy(st.lab2,lprt);
    strcat(st.lab2,"  ");
    strcat(st.lab2,st.comment);
    
    printf(" Reference Time      = %s\n", lprt);
    printf(" Pulsar Period (sec) = %f\n", runp.period);
        
    flag_plot_done=0;
    tinteg=0.0;
    p_usampl=0;
    p_adata=0;
    // initialize array
    for(k=0; k<4*st.nadata; k++) st.adata[k]=0.0;

    st.flag_g_device_open=0;

    ns_all=0;

    runp.npulse = 20;
    //runp.tspan = 20.0;
    tintegold = 0.0;
    npulse = 0;
    nloop=0;
loop1:
    if(nloop >0) {
        if((kbytes=CheckK5header_with_Read(&K5)) < 0){
            //i=fclose( K5.stream ) ;
            flag_last = 1;
            goto last_plot;
        }
        printf("  Time %02d:%02d:%02d   Sec in Day = %d\n",
            K5.hh,K5.mm,K5.ss,K5.seconds);
        //runp.dtime = (double)K5.seconds;
    }


    for(i=0; i<K5.imax; i++){
        K5.wordread = fread(K5.ubuf, sizeof(u_int32_t), K5.nword, K5.stream);
        if(K5.wordread < K5.nword){  // probably no more data     2015-10-27
            flag_last=1;
            if(K5.wordread==0){
                tusampl=0;
                goto exithere;
            }
            /* recalc tusampl and fact */
            if(st.adbit*st.numch >= 8){
                usampl=K5.wordread*4/(st.adbit*st.numch/8);
                fact=((float)usampl)*((float)usampl);
                tusampl=(double)usampl/K5.fsample;   // period of usampl
            } else {
                tusampl=0;
                goto exithere;
            }
        }

        K5.pos+=K5.wordread*4;
        K5.vflag=1; // validiate data
        parseADdata_3N(&K5);

        if(start_flag==0){
            if(ftime < fsoffset){
                ftime+=tusampl;
                continue;
            } else {
                start_flag=1;
            }
        }

        /* check current sample # */
        //p_usampl+=K5.npos/K5.numch;



        for(k=0; k<K5.npos/K5.numch; k++){
            kaisu=0;
            for(m=0; m<st.numch; m++){
                /* set data */
                if(p_adata < st.nadata){
                    if(p_adata==0 && m==0){  // keep time offset of begining of adata
                        ns_bopp=ns_all;
                        st.toffset=(double)ns_bopp*K5.tsample;
                    }
                    st.adata[p_adata*4+m]=(float)K5.idata[k*st.numch+m];


                } else {
                    // compute power  2020-05-20
                    if (st.flag_power == 1) {  // 2020-05-20
#ifdef MOVING_AVE
                        calc_power(st.adata, m, st.nadata, nave, 1);  // runnning average
#else
                        calc_power(st.adata, m, st.nadata, nave, 0);  // simple average
#endif
                    }
                    // 
                    if (m == 0) {   //  prepare for pulsar synvhronized ensemblle averaging
                        
                        for (n = 0; n < st.nadata; n++) {
                            dw = runp.dtime + (double)n*K5.tsample;
                            tseclast = dw;
                            dw -= runp.prtsec;
                            dw = fmod(dw / runp.period, 1.0) + 1.0;
                            dw = fmod(dw, 1.0)*runp.npdat;
                            npos = (int)(dw + 0.5);
                            if (npos >= runp.npdat) npos = runp.npdat - 1;
                            runp.pdat[npos] += st.adata[n * 4 + m];
                            runp.cdat[npos]++;
                        }
                    }
                    // plot here
                    //if( tinteg > 0.0 && ((int)tinteg % 10)==0)
                    //        plotpulse(&st, &runp);
            //        for(i=0; i<4*st.nadata; i++) st.adata[i]=0.0;
            //        flag_plot_done=1;
                    p_adata=0;
                    p_usampl=k;
                    //st.toffset+=st.tspan;
                    goto break1;
                }
                kaisu++;
            }

            p_adata++;
            ns_all++;
        break1:;
            //if (tinteg > 0.0 && ((int)tinteg % 10) == 0)
            //    plotpulse(&st, &runp);
            /*
            if(flag_plot_done==1){
                if(st.flag_halt==1){
                   //printf(" Hit return key for next span (CTRL-C for stop)  ");
                   printf(" Hit return key for next span (or \"N\"on-stop or \"Q\"uit) ");
                   cw=getchar();  // dummy read to accept just return 
                   if(cw=='N' || cw=='n') st.flag_halt=0;
                   if(cw=='Q' || cw=='q') goto exitloop;
                } else {
                   spe_sleep(st.sleep_msec);
                }
                flag_plot_done=0;
            }
            */
        }

exithere:
        tinteg+=tusampl;
        runp.dtime += tusampl;

        if (((int)tinteg-(int)tintegold)==1 && ((int)tinteg % 10) == 0) {
            runp.npulse = (tseclast - tsec1st) / runp.period;   // # of pulses
            strcpy(ltlast, sectohhmmss(tseclast));
            //sprintf(st.comment, "data span (sec) = %f", tinteg);
            sprintf(st.comment, "data span  %s-%s", lt1st,ltlast);
            strcpy(st.lab2, lprt);
            strcat(st.lab2, "  ");
            strcat(st.lab2, st.comment);
            plotpulse(&st, &runp);
            //flag_plot_done = 0;
        }
        tintegold = tinteg;
        if (runp.tlength > 0.0) {
            if (tinteg >= runp.tlength) flag_last = 1;
        }

    last_plot:;
        if(flag_last == 1){
            if(tinteg==0.0) goto exitloop;
            runp.npulse = (tseclast - tsec1st) / runp.period;   // # of pulses
            strcpy(ltlast, sectohhmmss(tseclast));
            //sprintf(st.comment, "data span (sec) = %f", tinteg);
            sprintf(st.comment, "data span  %s-%s", lt1st, ltlast);
            strcpy(st.lab2, lprt);
            strcat(st.lab2, "  ");
            strcat(st.lab2, st.comment);
            plotpulse(&st, &runp);
            goto exitloop;  
        }

    }
    nloop++;
    goto loop1;
exitloop:
    i=fclose( K5.stream ) ;

    cfinish1sec = clock();
    telapsed = (double)(cfinish1sec - cstart1sec) / CLOCKS_PER_SEC;
    //printf("Time elapsed for processing is %f sec\n",telapsed);

  exitend: ;
      
    free(K5.ubuf);
    free(K5.idata);
    free(st.adata);

 exitend2: ;

    return 0;
}



void plotpulse(speana_param_t *st, runp_t *rp)
{
    /*
           plot pulse
                 2020-05-26
    */
    int kplot, m, k, n;
    char labfsampl[10];

    //st->xdata = (float *)malloc(st->nadata * sizeof(float));
    //st->tdat = (float *)malloc(st->nadata * sizeof(float));
    st->xdata = (float *)malloc(rp->npdat * sizeof(float));
    st->tdat = (float *)malloc(rp->npdat * sizeof(float));



    /* prepare time axis data */
    //for (n = 0; n < st->nadata; n++) {
    for (n = 0; n < rp->npdat; n++) {
        //st->tdat[n] =rp->period*(double)n /(double)(rp->npdat-1);
        st->tdat[n] =360.0*(double)n /(double)(rp->npdat-1);
    }

    /* sampling frequency in text for print */
    if (st->sfreq < 1000) {
        sprintf(labfsampl, "%dkHz", st->sfreq);
    }
    else {
        sprintf(labfsampl, "%dMHz", st->sfreq / 1000);
    }

    for (kplot = 0; kplot < 2; kplot++) {
        if (st->pmode == 1 && kplot == 1) goto bypass;
        if (st->pmode == 2 && kplot == 0) goto bypass;
        if (kplot == 0) {
            strcpy(st->g_device, st->g_ps_device);
        }
        else {
            strcpy(st->g_device, st->g_display);
        }


        //if(kplot==0) {
        //    strcpy(pgdevice,"%s/PS",st->ps_fname);
        //} else {
        //    strcpy(pgdevice,st->g_device);
        //}
        if (st->flag_g_device_open == 0) {
            if (cpgopen(st->g_device) < 1) goto exitend;
            st->flag_g_device_open = 1;
        }
        if (strncmp(st->g_device, "/GW", 3) == 0) cpgeras();   // erase just in case 2007-10-26
        if (strncmp(st->g_device, "/XWIN", 5) == 0 ||
            strncmp(st->g_device, "/XSER", 5) == 0) { /* set background white
                                                    and text black */
            cpgscr(0, 1.0, 1.0, 1.0);  /* Background color */
            cpgscr(1, 0.0, 0.0, 0.0);  /* Foreground color */
        }

        if (st->numch != 1) {
            cpgsubp(2, 2);  // divide plot pannel into 2X2
        }
        for (m = 0; m < st->numch; m++) {


            /*
            if (st->tspan < 1.0) {
                if (st->tspan < 1.0e-3) {
                    sprintf(st->lab1, "CH#:%1d %1dbit %s sampling, T_span(usec)=%6.2f",
                        // m+1,st->adbit,labfsampl,st->tspan*1.0e6);
                        st->chtable[m], st->adbit, labfsampl, st->tspan*1.0e6);
                }
                else {
                    sprintf(st->lab1, "CH#:%1d %1dbit %s sampling, T_span(msec)=%.3f",
                        //m+1,st->adbit,labfsampl,st->tspan*1.0e3);
                        st->chtable[m], st->adbit, labfsampl, st->tspan*1.0e3);
                }
            }
            else {
                sprintf(st->lab1, "CH#:%1d %1dbit %s sampling, T_span(sec)=%.1f",
                    //m+1,st->adbit,labfsampl,st->tspan);
                    st->chtable[m], st->adbit, labfsampl, st->tspan);
            }
            */
            //sprintf(st->lab1, "CH#:%1d %1dbitX%s, Pulse_period(sec)=%f",
            sprintf(st->lab1, "CH#:%1d %1dbitX%s, Tp(sec)=%f, # of Pulses = %d",
                    //m+1,st->adbit,labfsampl,st->tspan);
                st->chtable[m], st->adbit, labfsampl, rp->period, rp->npulse);

            //for (k = 0; k < st->nadata; k++) {
            for (k = 0; k < rp->npdat; k++) {
                //st->xdata[k] = st->adata[k * 4 + m];
                if (rp->cdat[k] > 0) {
                    st->xdata[k] = rp->pdat[k] / (double)rp->cdat[k];
                }
                else {
                    st->xdata[k] = 0.0;
                }
            }

            plotpulse_a(st, rp, m);

        }

        /* panel undivided to date write */
        cpgsubp(1, 1);
        /* date etc write */
        cpgiden();
        /* data file name write */
        cpgsch(0.7f);                 // "f" means single precision floating number
        cpgsvp(0.0, 1.0, 0.0, 1.0);   // set viewport as all area
        cpgswin(0.0, 1.0, 0.0, 1.0);   // combine with world corrdinates
        cpgtext(0.01f, 0.01f, st->fname);
        /* close the graphics device */
        if (strncmp(st->g_device, "/GW", 3) != 0) {
            cpgclos();   //2007-10-26
            st->flag_g_device_open = 0;
        }
        if (kplot == 0) {
            printf("Created PostScript file ==> %s\n", st->ps_fname);
        }
    bypass:;
    }
exitend:;
    free(st->xdata);
    free(st->tdat);
}



void plotpulse_a(speana_param_t *st, runp_t *rp, int mch)
/*

           2017-12-08  revised to improve appearance in case of adbit=2
           2020-05-20  support power plot mode

               mode2 -- YX: axis scale mode
                          X: frequency axis mode
                             =0: log scale  =1: linear scale
                          Y: amplitude axis mode
                             =0: log scale
                             =1: linear scale (common for 4 channels)
                                  (use ymaxdb, ymindb)
                             =2: linear scale (each max and min)
*/
{
    float tmax, tmin, fsampl, amax, amin;
    float tmax0, tmin0, tmaxr, tminr;
    float *tdat, *adat;
    float thaba, tscale;
    float prange; // 2020-05-20
    double dw, dave, sigma;  // 2020-05-20
    char htitl[60], xtit[60], ytit[60];
    char lab2[128];
    char cw;
    int n, npmax;

    htitl[0] = '\0';
    //strcpy(htitl,"TIME SERIES");

    //npmax=st->usampl;
    npmax = rp->npdat;

    tdat = (float *)malloc(npmax * sizeof(float));
    adat = (float *)malloc(npmax * sizeof(float));

    /* time axis */
    fsampl = (float)1000.0*(float)st->sfreq;
    //thaba = (float)npmax / fsampl;
    tscale = 1.0;

    //strcpy(xtit, "TIME (sec)");
    strcpy(xtit, "Phase (deg)");
    //sprintf(lab2, "%s offset(msec)=%10.3f", st->lab2, st->toffset*1.0e3);
    sprintf(lab2, "%s", st->lab2);

    for (n = 0; n < npmax; n++) tdat[n] = st->tdat[n];
    tmax0 = tdat[npmax - 1];
    tmin0 = tdat[0];
    thaba = tmax0 - tmin0;

    //for (n = 0; n < npmax; n++) tdat[n] *= tscale;

    tmin = tmin0;
    tmax = tmax0;

//    if (tmax > st->tspan) tmax = (float)st->tspan;
    //tmax*=tscale;
    //tmin*=tscale;


    /* amplitude axis */

    /* data copy */
    for (n = 0; n < npmax; n++) adat[n] = st->xdata[n];

    /* data max, min check */
//      for(n=0; n<npmax; n++){
//         if(n==0){
//            amax=adat[n];
//            amin=adat[n];
//         } else {
//            if(adat[n] > amax) amax=adat[n];
//            if(adat[n] < amin) amin=adat[n];
//         }
//      }

    amin = 0.0;
    amax = (float)pow(2, (double)st->adbit);
    if (st->adbit == 1) {
        amin = (float)(-0.2);
        amax = (float)1.2;
    }
    else if (st->adbit == 2) {        // added on 2017-12-08
        amin = (float)(-0.2);         // added on 2017-12-08
        amax = (float)3.2;            // added on 2017-12-08
    }

    strcpy(ytit, "AMPLITUDE");

    if (st->flag_power == 1) {   // added on 2020-05-20
        strcpy(ytit, "RELATIVE POWER (dB)");
        // convert dB
        //dave = 0.0;
        for (n = 0; n < npmax; n++) {
            dw = 10.0*log10((double)adat[n]);
            adat[n] = (float)dw;
            //dave += dw;
        }
        dave = get_ave_sigma(adat, npmax, &sigma);   // 2020-05-20
        for (n = 0; n < npmax; n++) {
            adat[n] -= (float)dave;
            if (n == 0) {
                amax = adat[n];
                amin = adat[n];
            }
            else {
                if (adat[n] > amax) amax = adat[n];
                if (adat[n] < amin) amin = adat[n];
            }
        }

        if (st->minpow != 0.0 || st->maxpow != 0.0) {
            amin = (float)st->minpow;
            amax = (float)st->maxpow;
        }
        else { // auto mode
            prange = amax - amin;
            amax += 0.1*prange;
            amin -= 0.1*prange;
        }
    }
loop:
    tmaxr = tmax * tscale;
    tminr = tmin * tscale;




    cpgsch(2.0);  // set text height 2.0  (default is 1)
    cpgslw(2);    // set line width

    cpgpanl((mch % 2) + 1, (mch / 2 + 1));  // when use this, should not use cpgenv
                                         //    mch    ix      iy        cpgpanl(ix,iy)
                                         //     0      1       1
                                         //     1      2       1
                                         //     2      1       2
                                         //     3      2       2

    cpgeras();  // doesn't work properly with cpgenv

    /* define coordinate range of graph */
//      cpgenv(tminr, tmaxr, amin, amax, 0, 0);   // <= this doesn't work well when cpgpanl is called
                                                  //

      //cpgsvp(0.4,0.9,0.2,0.2);  // this doesn't work for sub panel
    cpgvstd();
    cpgswin(tminr, tmaxr, amin, amax);          //  these two lines work as same as cpgenv()
    //cpgbox("BCNTS",0.0,0,"BCNTSV",0.0,0);        //
    cpgbox("BCNTS", 90.0, 3, "BCNTS", 0.0, 0);        //


    /* label the axes */
    cpglab(xtit, ytit, htitl);

    /* sub label write*/
    cpgsch(1.5);
    cpgmtxt("T", 1.5, 0.5, 0.5, st->lab1);
    //cpgmtxt("T",0.5,0.5,0.5,st->lab2);
    cpgmtxt("T", 0.5, 0.5, 0.5, lab2);
    cpgsch(2.0);

    /* check data */
    if (amax == amin) {
        cpgmtxt("B", -6.0, 0.5, 0.5, "NO TIME VARIABLE");
        cpgmtxt("B", -4.0, 0.5, 0.5, "COMPONENT FOUND");
        /* label the axes */

    }
    else {

        /* plot the line graph */
        tdat[0] = tdat[1];
        adat[0] = adat[1];  // to avoid ugry plot
        //cpgslw(1);    // set line width
        cpgline(npmax, tdat, adat);

    }

/*
    if (tmax < tmax0) {
        tmin = tmax;
        tmax += (float)st->tspan;
        if (st->flag_halt == 1) {
            printf(" Hit return key for next span ");
            cw = getchar();  // dummy read to accept just return 
        }
        else {
            spe_sleep(st->sleep_msec);
        }
        goto loop;
    }
*/
    free(adat);
    free(tdat);
}

char *sectohhmmss(double dsec)
{
/*
     seconds to HH:MM:SS.SSS
     2020-05-27
*/
    static char cbuf[20];
    int hh, mm;
    double ss;

    hh = (int)(dsec / 3600.0);
    mm = (int)(fmod(dsec, 3600.0) / 60.0);
    ss = fmod(dsec, 60.0);
    sprintf(cbuf, "%02d:%02d:%06.3f", hh, mm, ss);
    return cbuf;
}